PetInfoFu = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0",
										   "AceConsole-2.0",
										   "AceDB-2.0",
										   "AceEvent-2.0"
										  )

 -- variables
local L          = AceLibrary("AceLocale-2.2"):new("PetInfoFu")
local BST        = AceLibrary("Babble-SpellTree-2.2")
local crayon     = AceLibrary("LibCrayon-3.0")
local dewdrop    = AceLibrary("Dewdrop-2.0")
local tablet     = AceLibrary("Tablet-2.0")

-- hunter specific variables 

-- warlock specific variables
local demonicKnowledgeRank = 0
local demonicKnowledgeTalentTab = 0
local demonicKnowledgeTalentIndex = 0

-- properties
PetInfoFu.clickableTooltip = true
PetInfoFu.canHideText      = true
PetInfoFu.defaultPosition  = "RIGHT"
PetInfoFu.hasIcon          = true
PetInfoFu.hasNoColor       = false

-- register database stuff
PetInfoFu:RegisterDB("PetInfoFuDB")
PetInfoFu:RegisterDefaults("account", {
										stables = {},
									  }
						  )
						  
PetInfoFu:RegisterDefaults("char", {
									-- flags
									showPetName = true,
									showAttributes = true,
									showCombatStats = true,
									showSkills = true,
					showActivePets = true,
									showStabledPets = false,
									showMendPetFrame = false,
									-- sound play
									-- frame locations
									petInfoFramePoint = { point = "CENTER", x = 0, y = 0},
									petMendFramePoint = { point = "CENTER", x = 0, y = 0},
									-- lists
									hidden = {},
					activePets = {}, 	    -- stores first 5 pets
									stabledPets = {},       -- stores currently stabled pets (6,25)
									petInfo = {},           -- stores pet skills and similar info
									netheredDemons = {},    -- stores info about all demons
								   } 
						 )
						 
PetInfoFu:RegisterChatCommand(L["CMD_OPTIONS"])

-- menu methods
function PetInfoFu:IsShowingPetName()
	return self.db.char.showPetName
end

function PetInfoFu:ToggleIsShowingPetName()
	self.db.char.showPetName = not self.db.char.showPetName
	if not (self.db.char.showPetName or self.db.char.showPetLevel) then self:ShowIcon() end
	self:UpdateText()
	return self.db.char.showPetName
end

function PetInfoFu:IsShowingAttributes()
	return self.db.char.showAttributes
end

function PetInfoFu:ToggleIsShowingAttributes()
	self.db.char.showAttributes = not self.db.char.showAttributes
	self:UpdateTooltip()
	return self.db.char.showAttributes
end

function PetInfoFu:IsShowingCombatStats()
	return self.db.char.showCombatStats
end

function PetInfoFu:ToggleIsShowingCombatStats()
	self.db.char.showCombatStats = not self.db.char.showCombatStats
	self:UpdateTooltip()
	return self.db.char.showCombatStats
end

function PetInfoFu:IsShowingSkills()
	return self.db.char.showSkills
end

function PetInfoFu:ToggleIsShowingSkills()
	self.db.char.showSkills = not self.db.char.showSkills
	self:UpdateTooltip()
	return self.db.char.showSkills
end

function PetInfoFu:IsShowingMendPetFrame()
	return self.db.char.showMendPetFrame
end

function PetInfoFu:ToggleIsShowingMendPetFrame()
	self.db.char.showMendPetFrame = not self.db.char.showMendPetFrame
		if not self.db.char.showMendPetFrame and self.petMendFrame then
			self.petMendFrame:Hide()
		end 
	return self.db.char.showMendPetFrame
end

function PetInfoFu:IsShowingStabledPets()
	return self.db.char.showStabledPets
end

function PetInfoFu:ToggleIsShowingStabledPets()
	self.db.char.showStabledPets = not self.db.char.showStabledPets
	self:UpdateTooltip()
	return self.db.char.showStabledPets
end

function PetInfoFu:IsShowingActivePets()
	return self.db.char.showActivePets
end

function PetInfoFu:ToggleIsShowingActivePets()
	self.db.char.showActivePets = not self.db.char.showActivePets
	self:UpdateTooltip()
	return self.db.char.showActivePets
end

function PetInfoFu:IsTrackingPetSkills()
	return self.db.char.trackPetSkills
end

function PetInfoFu:ToggleIsTrackingPetSkills()
	self.db.char.trackPetSkills = not self.db.char.trackPetSkills
	self:UpdateText()
	return self.db.char.trackPetSkills
end

-- sound toggles
   
-- init/exit functions
function PetInfoFu:OnInitialize()
  --nothing here
end

function PetInfoFu:OnEnable()
	-- update
	self:RegisterEvent("UNIT_NAME_UPDATE","Update")
	self:RegisterEvent("UNIT_PET")
	-- Fires when the player's pet is dismissed. Applies to warlock minions, mind controlled units, quest-related pets, etc., 
	-- but not to hunter pets
	-- self:RegisterEvent("PET_DISMISS_START")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	-- local event handlers
	if self:IsHunter() then	
		self:RegisterEvent("PET_STABLE_CLOSED")
		-- combat log tracking
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		--      
		self:ScheduleRepeatingEvent(self.name, self.UpdateText, 60, self)         
	end

	if self:IsWarlock() then
		self:GetDemonicKnowledgeRank()
		self:RegisterEvent("CHARACTER_POINTS_CHANGED")
	end

	self:SetIcon(select(3, GetSpellInfo(2641)))  -- Dismiss Pet icon

end

function PetInfoFu:OnDisable()
	self:UnregisterAllEvents()
end

-- events (sorted alphabetically)
function PetInfoFu:CHARACTER_POINTS_CHANGED()
	self:GetDemonicKnowledgeRank()
end

function PetInfoFu:COMBAT_LOG_EVENT_UNFILTERED(_, eventType, hideCaster, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType, ...)
	if spellId ==  2641 then     -- Dismiss Pet
		self:StoreCurrentPet()		
	end
	-- spellId ==  883   -- Call Pet 1
	-- spellId ==  83242 -- Call Pet 2	
	-- spellId ==  83243 -- Call Pet 3
	-- spellId ==  83244 -- Call Pet 4
	-- spellId ==  83245 -- Call Pet 5
end

function PetInfoFu:PET_STABLE_CLOSED()
	self:GetActivePets()
	self:GetStabledPets()
	-- self:Print('PetInfoFu: Tooltip'..crayon:Red('PET_STABLE_CLOSED'))
end

function PetInfoFu:PLAYER_ENTERING_WORLD()
	self:GetActivePets()
	self:GetStabledPets()
	-- self:Print('PetInfoFu: Tooltip'..crayon:Red('PLAYER_ENTERING_WORLD'))
end

function PetInfoFu:UNIT_PET()
	self:GetActivePets()
	self:GetStabledPets()
	-- self:Print('PetInfoFu: Tooltip'..crayon:Red('UNIT_PET'))
end

-- frame functions
function PetInfoFu:PetMendFrame()
	if self.petMendFrame ~= nil then 
		self.petMendFrame:Show()
		return
	end	
	
	-- create frame
	local frame = CreateFrame("Frame", "PetInfoFuMendFrame", UIParent)
	frame:SetPoint(self.db.char.petMendFramePoint.point, self.db.char.petMendFramePoint.x, self.db.char.petMendFramePoint.y)
	frame:SetFrameStrata("TOOLTIP")
	frame:SetFrameLevel(99)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function(this)
		this:StartMoving()
	end)
	
	frame:SetScript("OnDragStop", function(this)
		this:StopMovingOrSizing()
		local point,relativeTo,relativePoint,xOfs,yOfs = this:GetPoint()
		self.db.char.petMendFramePoint.point = point
		self.db.char.petMendFramePoint.x = xOfs
		self.db.char.petMendFramePoint.y = yOfs
	end)

	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	})
	frame:SetBackdropBorderColor(.5, .5, .5, 1)
	frame:SetBackdropColor(0, 0, 0, .3)
	
	-- create heal icon
	local spellIcon = frame:CreateTexture(nil, "ARTWORK")
	frame.spellIcon = spellIcon
	spellIcon:SetTexture(select(3, GetSpellInfo(136)))
	spellIcon:SetWidth(26)
	spellIcon:SetHeight(26)
	spellIcon:SetPoint("TOPLEFT", frame, "TOPLEFT", 7, -7)

	-- activate frame
	frame.name = UnitName("pet")
	self.petMendFrame = frame
	frame:SetWidth(40)
	frame:SetHeight(40)
	frame:Show()    
end

function PetInfoFu:PetInfoFrame(petInfo)
	-- with kind thanks to TrainerFu for the basic routine
	
	-- hide other frames, destroy if same frame is called again
	if self.petInfoFrame then 
		self.petInfoFrame:Hide()
		if self.petInfoFrame.name == petInfo.PetName then
			self.petInfoFrame = nil
			return
		end
	end	
   
	local text = crayon:Green(petInfo.PetName..' - '..petInfo.PetFamily..' ('..petInfo.PetTalent..')\n')
	
	local petSaveKey = petInfo.PetFamily.."|"..petInfo.PetName
	-- if available add list of pet skills
	if self.db.char.petInfo[petSaveKey] ~= nil then
		if self.db.char.petInfo[petSaveKey]['petSkills'] ~= nil then
			text = text..'\n'
			-- create a sorted table of pet skills
			sortedPetSkills = {}
			for k,v in pairs(self.db.char.petInfo[petSaveKey]['petSkills']) do
				-- self:Print('PetInfoFu: Adding'..crayon:Red(k)..' '..v.SpellRank)
				if v.SpellRank == '' then
					table.insert(sortedPetSkills, k)                
				else
					table.insert(sortedPetSkills, k..' - '..v.SpellRank)
				end
			end
			table.sort(sortedPetSkills, function (a, b) return a < b end)
			-- list sorted pet skills
			for i = 1, #sortedPetSkills do  
				text = text..'\n'..sortedPetSkills[i]
			end
		 end
	end
	
	-- create frame
	local frame = CreateFrame("Frame", "PetInfoFuPetFrame", UIParent)
	frame:SetPoint(self.db.char.petInfoFramePoint.point, self.db.char.petInfoFramePoint.x, self.db.char.petInfoFramePoint.y)
	frame:SetFrameStrata("TOOLTIP")
	frame:SetFrameLevel(99)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetScript("OnDragStart", function(this)
		this:StartMoving()
	end)
	frame:SetScript("OnDragStop", function(this)
		this:StopMovingOrSizing()
		local point,relativeTo,relativePoint,xOfs,yOfs = this:GetPoint()
		self.db.char.petInfoFramePoint.point = point
		self.db.char.petInfoFramePoint.x = xOfs
		self.db.char.petInfoFramePoint.y = yOfs
	end)

	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4},
	})
	frame:SetBackdropBorderColor(.5, .5, .5, 1)
	frame:SetBackdropColor(0, 0, 0, .3)
	
	-- create pet icon
	if self.db.char.petInfo[petSaveKey] ~= nil then
	if self.db.char.petInfo[petSaveKey].PetIcon ~= nil then        
		local petIcon = frame:CreateTexture(nil, "ARTWORK")
		frame.petIcon = petIcon
		petIcon:SetTexture(self.db.char.petInfo[petSaveKey].PetIcon)
		petIcon:SetWidth(40)
		petIcon:SetHeight(40)
		petIcon:SetPoint("TOPRIGHT", frame, "TOPLEFT", 1, -3)
	end
	end

	-- create close button
	local closeButton = CreateFrame("Button", "PetInfoFuCloseButton", frame)
	closeButton:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0)
	closeButton:SetWidth(30)
	closeButton:SetHeight(30)
	closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
	closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down")
	closeButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight")
	closeButton:SetScript("OnClick", function()
		frame:Hide()
	end)

	-- set actual tooltip text
	local petText = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	frame.petText = petText
	petText:SetText(text)
	petText:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10)
	petText:SetWidth(280)
	petText:SetJustifyH("LEFT")
	petText:SetJustifyV("TOP")

	-- activate frame
	frame.name = petInfo.petName
	self.petInfoFrame = frame
	frame:SetWidth(petText:GetWidth() + 20)
	frame:SetHeight(petText:GetHeight() + 20)
	frame:Show()    
end

-- functions 

function PetInfoFu:StoreCurrentPet()
	local petName = UnitName("pet")    
	local petFamily = UnitCreatureFamily("pet")

	if petName ~= nil and 
	   petName ~= UNKNOWNOBJECT then    
	   
		local petSaveKey = petFamily.."|"..petName    
		-- init list item if not exists
		if self.db.char.petInfo[petSaveKey] == nil then
			self.db.char.petInfo[petSaveKey] = {}        
		end
		self.db.char.petInfo[petSaveKey] = {
			PetName = petName,
			PetFamily = petFamily,
			PetLevel = UnitLevel("pet"),
			PetIcon = GetPetIcon()
		}
		local petName = UnitName("pet") 
		self:Print('PetInfoFu: Storing '..crayon:Red(petName))	
		
		-- list pet skills, reset curent list
		self.db.char.petInfo[petSaveKey]['petSkills'] = {}
		local i = 1
		while true do
			local spellName, spellRank = GetSpellInfo(i, BOOKTYPE_PET)
			-- self:Print('PetInfoFu: spell '..crayon:Red(spellName))	
			
			if not spellName then
				do break end
			end
			self.db.char.petInfo[petSaveKey]['petSkills'][spellName]  = {
				SpellRank = spellRank
			}
			i = i + 1
		end
	end
end

function PetInfoFu:GetTalentColour(petTalent)
	if petTalent == L["PI_TALENT_FEROCITY"] then return crayon:Colorize('fff555',petTalent) end
	if petTalent == L["PI_TALENT_CUNNING"] then return crayon:Colorize('55f5ff',petTalent) end
	if petTalent == L["PI_TALENT_TENACITY"] then return crayon:Colorize('ff5533',petTalent) end
end

function PetInfoFu:GetActivePets()
	self.db.char.activePets = nil
	self.db.char.activePets = {}
	for i = 1,5 do -- active pets
		local _, petName, _, petFamily, petTalent = GetStablePetInfo(i, true)
		if petName ~= nil then
			-- self:Print('PetInfoFu: Active '..crayon:Red(petName)..' '..format(i))         
			self.db.char.activePets[i] = {
				PetName = petName,
				PetFamily = petFamily,
		PetTalent = petTalent
			}
		end
	end
	self:Update()
end

function PetInfoFu:GetStabledPets()
	self.db.char.stabledPets = nil
	self.db.char.stabledPets = {}
	for i = 6,25 do -- stabled pets
		local _, petName, _, petFamily, petTalent = GetStablePetInfo(i, true)
		if petName ~= nil then
		   -- self:Print('PetInfoFu: Stable '..crayon:Red(petName)..' '..format(i))         
			self.db.char.stabledPets[i] = {
				PetName = petName,
				PetFamily = petFamily,
		PetTalent = petTalent
			}
		end
	end
	self:Update()
end

function PetInfoFu:GetDemonicKnowledgeRank()   
	 if (demonicKnowledgeTalentTab == 0 or demonicKnowledgeTalentIndex == 0) then
		for i = 1, GetNumTalentTabs() do
			local tabName = GetTalentTabInfo(i)
			if (tabName == BST["Demonology"]) then
				demonicKnowledgeTalentTab = i
				for j = 1, GetNumTalents(i) do
					local talentName, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, j)
					if talentName == GetSpellInfo(36691) then   -- Demonic Knowledge
						demonicKnowledgeTalentIndex = j
						--set currentRank
						demonicKnowledgeRank = currentRank
						
					end
				end
			end
		end
	else
		local _, _, _, _, currentRank = GetTalentInfo(demonicKnowledgeTalentTab, demonicKnowledgeTalentIndex)
		-- set currentRank
		demonicKnowledgeRank = currentRank
	end
	self:Update()
end

function PetInfoFu:UpdateNetheredDemon(petName,petType,spellName,spellRank)
	if petName ~= nil then   
		if self.db.char.netheredDemons[petName] == nil then
			self.db.char.netheredDemons[petName] = {}
		end
		self.db.char.netheredDemons[petName].PetName = petName
		self.db.char.netheredDemons[petName].PetType = petType
		if self.db.char.netheredDemons[petName][spellName] == nil then
			self.db.char.netheredDemons[petName][spellName] = {}
		end
		self.db.char.netheredDemons[petName][spellName] = {
			SpellName = spellName,
			SpellRank = spellRank
		}      
	end
end

function PetInfoFu:ToggleCategory(id, button)
	-- hidden is character specific data
	self.db.char.hidden[id] = not self.db.char.hidden[id]
	-- refresh in place
	self:UpdateTooltip()
end    
	   
-- is class functions
	
function PetInfoFu:IsHunter()
	local _, englishClass = UnitClass("player")
	return englishClass == "HUNTER"
end
	
function PetInfoFu:IsWarlock()
	local _, englishClass = UnitClass("player")
	return englishClass == "WARLOCK" 
end

-- menus
	
function PetInfoFu:OnMenuRequest(level, value)
	if level == 1 then
		-- label options
		dewdrop:AddLine(
			'text', L["Show pet's name"],
			'func', 'ToggleIsShowingPetName',
			'arg1', self,
			'closeWhenClicked', true,                
			'checked', self:IsShowingPetName()
		)
		-- tooltip options
		dewdrop:AddLine(
			'text', L["Show pet's attributes"],
			'func', 'ToggleIsShowingAttributes',
			'arg1', self,
			'closeWhenClicked', true,                
			'checked', self:IsShowingAttributes()
		)
		dewdrop:AddLine(
			'text', L["Show pet's combat statistics"],
			'func', 'ToggleIsShowingCombatStats',
			'arg1', self,
			'closeWhenClicked', true,                
			'checked', self:IsShowingCombatStats()
		)
		dewdrop:AddLine(
			'text', L["Show pet's skills"],
			'func', 'ToggleIsShowingSkills',
			'arg1', self,
			'closeWhenClicked', true,                
			'checked', self:IsShowingSkills()
		)
	   if self:IsHunter() then
			dewdrop:AddLine(
				'text', L["Show Mend Pet frame"],
				'func', 'ToggleIsShowingMendPetFrame',
				'arg1', self,
				'closeWhenClicked', true,                
				'checked', self:IsShowingMendPetFrame()
			)
			dewdrop:AddLine(
				'text', L["CATEGORY_PET_STABLED"],  -- stub for popup
				'arg1', self,
				'hasArrow', true,
				'value', "STABLEDPETS"
			)        
		end
		if self:IsWarlock() then
			dewdrop:AddLine(
				'text', L["Show nethered demons"],
				'func', 'ToggleIsShowingStabledPets',
				'arg1', self,
				'closeWhenClicked', true,         
				'checked', self:IsShowingStabledPets()
			)        
		end        
		dewdrop:AddLine()
	end
	if level == 2 then
		if value == "STABLEDPETS" then
			dewdrop:AddLine(
				'text', L["Show active pets"],
				'func', 'ToggleIsShowingStabledPets',
				'arg1', self,
				'closeWhenClicked', true,         
				'checked', self:IsShowingActivePets()
			)
			dewdrop:AddLine(
				'text', L["Show stabled pets"],
				'func', 'ToggleIsShowingStabledPets',
				'arg1', self,
				'closeWhenClicked', true,         
				'checked', self:IsShowingStabledPets()
			)
			dewdrop:AddLine(
				'text', L["Track pet skills"],
				'func', 'ToggleIsTrackingPetSkills',
				'arg1', self,
				'closeWhenClicked', true,                
				'checked', self:IsTrackingPetSkills()
			)            
		end
	end
end
	
-- updates    

function PetInfoFu:AddActivePets()
	local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_ACTIVE"], 'columns', 2,
										   'text', L["CATEGORY_PET_ACTIVE"],
										   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_ACTIVE"],
										   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
										   'showWithoutChildren', true,
										   'checked', true, 'hasCheck', true, 'checkIcon',
										   self.db.char.hidden[L["CATEGORY_PET_ACTIVE"]] and 
										   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
	)        
	if not self.db.char.hidden[L["CATEGORY_PET_ACTIVE"]] then
		for k, v in pairs(self.db.char.activePets) do  
			tooltipLine:AddLine(
				'text', v.PetName,
				'text2', v.PetFamily..' ('..self:GetTalentColour(v.PetTalent)..')',
				'text2R', 1, 'text2G', 1, 'text2B', 0,
				'func', 'PetInfoFrame', 'arg1', self, 'arg2', v
			)              
		end
	end
end

function PetInfoFu:AddStabledPets()
	local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_STABLED"], 'columns', 2,
										   'text', L["CATEGORY_PET_STABLED"],
										   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_STABLED"],
										   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
										   'showWithoutChildren', true,
										   'checked', true, 'hasCheck', true, 'checkIcon',
										   self.db.char.hidden[L["CATEGORY_PET_STABLED"]] and 
										   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
	)        
	if not self.db.char.hidden[L["CATEGORY_PET_STABLED"]] then
		for k, v in pairs(self.db.char.stabledPets) do  
			tooltipLine:AddLine(
				'text', v.PetName,
				'text2', v.PetFamily..' ('..self:GetTalentColour(v.PetTalent)..')',
				'text2R', 1, 'text2G', 1, 'text2B', 0,
				'func', 'PetInfoFrame', 'arg1', self, 'arg2', v
			)              
		end
	end
end
	
function PetInfoFu:ShowSkills()
	local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_SKILLS"], 'columns', 2,
										   'text', L["CATEGORY_PET_SKILLS"],
										   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_SKILLS"],
										   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
										   'showWithoutChildren', true,
										   'checked', true, 'hasCheck', true, 'checkIcon',
										   self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] and 
										   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
	)        
	if not self.db.char.hidden[L["CATEGORY_PET_SKILLS"]] then
		local petName = UnitName("pet")    
		local petType = UnitCreatureFamily("pet")
		local i = 1
		while true do
			local spellName, spellRank = GetSpellInfo(i, BOOKTYPE_PET)
			if not spellName then
				do break end
			end
			if self:IsWarlock() then
				self:UpdateNetheredDemon(petName,petType,spellName,spellRank)
			end
			tooltipLine:AddLine(
				'text', spellName,
				'text2', spellRank,
				'text2R', 1, 'text2G', 1, 'text2B', 0
			)
			i = i + 1
		end
	 end
end     

function PetInfoFu:AddNetheredDemons()
	local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_NETHER"], 'columns', 2,
										   'text', L["CATEGORY_PET_NETHER"],
										   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_NETHER"],
										   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
										   'child_indentation', 20,                                           
										   'showWithoutChildren', true,
										   'checked', true, 'hasCheck', true, 'checkIcon',
										   self.db.char.hidden[L["CATEGORY_PET_NETHER"]] and 
										   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
	)        
	
	if not self.db.char.hidden[L["CATEGORY_PET_NETHER"]] then
		-- create a sorted table of pets
		local sortedPets = {}
		for k,v in pairs(self.db.char.netheredDemons) do
			 table.insert(sortedPets, v)
		end
		table.sort(sortedPets, function (a, b) return a.PetType < b.PetType end)
		-- list sorted pets
		for i = 1, #sortedPets do
			if sortedPets[i].PetName ~= UnitName('pet') then
				local tooltipLine = tablet:AddCategory('id', sortedPets[i].PetName, 'columns', 2,
													   'text', sortedPets[i].PetName..' ('..sortedPets[i].PetType..')',
													   'func', 'ToggleCategory', 'arg1', self, 'arg2', sortedPets[i].PetName,
													   'indentation', 10,
													   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
													   'child_indentation', 25,
													   'showWithoutChildren', true,
													   'hideBlankLine', true,                                           
													   'checked', true, 'hasCheck', true, 'checkIcon',
													   self.db.char.hidden[sortedPets[i].PetName] and 
													   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
				)        
				if not self.db.char.hidden[sortedPets[i].PetName] then
					-- create a sorted table of skills
					local sortedSkills = {}
					for k,v in pairs(self.db.char.netheredDemons[sortedPets[i].PetName]) do
						if v.SpellName ~= nil then 
							table.insert(sortedSkills, v)
						end
					end
					table.sort(sortedSkills, function (a, b) return a.SpellName < b.SpellName end)
					-- list sorted skills
					for j = 1, #sortedSkills do
						tooltipLine:AddLine(
							'text', sortedSkills[j].SpellName,
							'text2', sortedSkills[j].SpellRank,
							'text2R', 1, 'text2G', 1, 'text2B', 0
						) 
					end                    
				end
			end
		end
	end
end
 
function PetInfoFu:OnTooltipUpdate()

	if UnitExists("pet") then

	tablet:SetHint(L["PI_TOOLTIP_HINT"])       

		
		local toolTip = tablet:AddCategory(
			'columns', 2,
			'child_textR', 0,
			'child_textG', 1,
			'child_textB', 0,
			'child_text2R', 1,
			'child_text2G', 1,
			'child_text2B', 0
		)
		
		local petName = UnitName("pet")
		toolTip:AddLine(
			'text', L["PI_PETNAME"],
			'text2', petName
		)        
		local petLevel = UnitLevel("pet")
		toolTip:AddLine(
			'text', L["PI_PETLEVEL"],
			'text2', petLevel
		)
		if self:IsHunter() then        
			-- general pet info            
			local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_INFO"], 'columns', 2,
												   'text', L["CATEGORY_PET_INFO"],
												   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_INFO"],
												   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
												   'showWithoutChildren', true,
												   'checked', true, 'hasCheck', true, 'checkIcon',
												   self.db.char.hidden[L["CATEGORY_PET_INFO"]] and 
												   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
			)
			if not self.db.char.hidden[L["CATEGORY_PET_INFO"]] then
			
				local petType = UnitCreatureFamily("pet")
				tooltipLine:AddLine(
					'text', L["PI_TYPE"],
					'text2', petType,
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
		
		local talent = GetPetTalentTree()
				tooltipLine:AddLine(
					'text', L["PI_TALENT"],
					'text2', talent,
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
				tooltipLine:AddLine(
					'text', L["PI_FOOD"],
					'text2', BuildListString(GetPetFoodTypes()),
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
				tooltipLine:AddLine() tooltipLine:AddLine() tooltipLine:AddLine()               
				-- training info
				-- update for talent points remaining
		points = GetUnspentTalentPoints(false, true, nil)
			   tooltipLine:AddLine(
					'text', L["PI_TRAINING_POINTS"], 
					'text2', points,
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
			end
		end
		
		if self:IsWarlock() then
			-- demonic knowledge
			-- skip section if not talented in demonic knowledge
			if demonicKnowledgeRank > 0 then
				local tooltipLine = tablet:AddCategory('id', L["CATEGORY_DEMONIC_KNOWLEDGE"], 'columns', 2,
													   'text', L["CATEGORY_DEMONIC_KNOWLEDGE"],
													   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_DEMONIC_KNOWLEDGE"],                                                   
													   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
													   'showWithoutChildren', true,
													   'checked', true, 'hasCheck', true, 'checkIcon',
													   self.db.char.hidden[L["CATEGORY_DEMONIC_KNOWLEDGE"]] and 
													   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'                                                  
				)
			
				if UnitExists("pet") then
					local demonSta, demonInt
					_, demonSta = UnitStat("pet", 3) -- stamina
					_, demonInt = UnitStat("pet", 4) -- intelligence
					tooltipLine:AddLine(
						'text', L["PI_BONUSDAMAGE"]..' ('..demonicKnowledgeRank..')',
						'text2', floor((demonSta + demonInt) * 0.05 * demonicKnowledgeRank),
						'text2R', 1, 'text2G', 1, 'text2B', 0
					)
				else
					tooltipLine:AddLine(
						'text', L["PI_NODEMON"]
					)            
				end
			end
		end
		
		-- combat info
		if self.db.char.showCombatStats then        
			local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_COMBAT_INFO"], 'columns', 2,
										   'text', L["CATEGORY_PET_COMBAT_INFO"],
										   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_COMBAT_INFO"],
										   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
										   'showWithoutChildren', true,
										   'checked', true, 'hasCheck', true, 'checkIcon',
										   self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] and 
										   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
			)
			if not self.db.char.hidden[L["CATEGORY_PET_COMBAT_INFO"]] then

				local baseArmor, effectiveArmor, armor, posArmorBuff, negArmorBuff = UnitArmor("pet");
				if posArmorBuff - negArmorBuff < 0 then
					effectiveArmor = crayon:Red(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Red("-"..format(baseArmor - effectiveArmor))..')'
				elseif baseArmor < effectiveArmor then 
					effectiveArmor = crayon:Green(format(effectiveArmor))..' ('..format(baseArmor)..crayon:Green("+"..format(effectiveArmor - baseArmor))..')'
				end
-- self:Print(baseArmor.." "..effectiveArmor.." "..armor.." "..posArmorBuff.." ".." "..negArmorBuff)                

				tooltipLine:AddLine(
					'text', L["PI_ARMOR_CLASS"], 
					'text2', effectiveArmor,
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
			  
				local atkBase, atkMod = UnitAttackBothHands("pet");        
				tooltipLine:AddLine(
					'text', L["PI_ATTACK_SKILL"],
					'text2', atkBase,
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)

				local atkPowerbase, atkPowerPosBuff, atkPowerNegBuff = UnitAttackPower("pet");
				if atkPowerPosBuff - atkPowerNegBuff < 0 then
					atkPowerbase = crayon:Red(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Red("-"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
				elseif atkPowerPosBuff - atkPowerNegBuff > 0 then
					atkPowerbase = crayon:Green(format(atkPowerbase + atkPowerPosBuff + atkPowerNegBuff))..' ('..format(atkPowerbase)..crayon:Green("+"..format(atkPowerPosBuff + atkPowerNegBuff))..')'
				end
				tooltipLine:AddLine(
					'text', L["PI_ATTACK_POWER"],
					'text2', atkPowerbase,
					'text2R', 1, 'text2G', 1, 'text2B', 0                
				)
				
				local atkSpeed = UnitAttackSpeed("pet")       
				tooltipLine:AddLine(
					'text', L["PI_ATTACK_SPEED"], 
					'text2', format(L["PI_AS_FORMAT"], atkSpeed),
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)

				local lowDmg, hiDmg, offlowDmg, offhiDmg, posBuff, negBuff, percentMod = UnitDamage("pet");
				local listedDPS = (((lowDmg + hiDmg) * .5 + posBuff + negBuff) * percentMod) / atkSpeed;
				tooltipLine:AddLine(
					'text', L["PI_DPS_LISTED"], 
					'text2', format(L["PI_DPS_FORMAT"], listedDPS),
					'text2R', 1, 'text2G', 1, 'text2B', 0
				)
			end
		end
		
		-- show stats info if showing
		if self.db.char.showAttributes then        
			local tooltipLine = tablet:AddCategory('id', L["CATEGORY_PET_ATTRIBUTES"], 'columns', 2,
												   'text', L["CATEGORY_PET_ATTRIBUTES"],
												   'func', 'ToggleCategory', 'arg1', self, 'arg2', L["CATEGORY_PET_ATTRIBUTES"],
												   'child_textR', 0, 'child_textG', 1, 'child_textB', 0,
												   'showWithoutChildren', true,
												   'checked', true, 'hasCheck', true, 'checkIcon',
												   self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] and 
												   'Interface\\Buttons\\UI-PlusButton-Up' or 'Interface\\Buttons\\UI-MinusButton-Up'
			)        
			if not self.db.char.hidden[L["CATEGORY_PET_ATTRIBUTES"]] then         
				for i = 1, NUM_PET_STATS, 1 do                  
					local attrbStr = L["SPELL_STAT"..format(i-1).."_NAME"]                   
					local baseStat, effectiveStat, posBuff, negBuff = UnitStat("pet", i)
			
					if posBuff == 0 and negBuff == 0 then
						baseStat = effectiveStat
					else 
						-- if there are any negative buffs then show the main number in red even if there are
						-- positive buffs. otherwise show in green.
						if negBuff < 0 then
							baseStat = crayon:Red(baseStat)..' ('..baseStat-negBuff..""..crayon:Red(negBuff+posBuff)..')'
						else
							baseStat = crayon:Green(effectiveStat)..' ('..baseStat-posBuff.."+"..crayon:Green(posBuff)..')'
						end
					end
					tooltipLine:AddLine(
						'text', attrbStr,
						'text2', baseStat,
						'text2R', 1, 'text2G', 1, 'text2B', 0
					)
				end
			end
		end
		
		-- show skills info
		if self.db.char.showSkills then          
			self:ShowSkills()
		end

		-- add other demons if showing
		if self:IsWarlock() and self.db.char.showStabledPets then
			self:AddNetheredDemons()
		end   
		
		-- add other pets if showing
		if self:IsHunter() and self.db.char.showActivePets then          
			self:AddActivePets()
		end   

		-- add other pets if showing
		if self:IsHunter() and self.db.char.showStabledPets then          
			self:AddStabledPets()
		end   
	else
		-- no pet found
		local toolTip = tablet:AddCategory(
			'columns', 1,
			'child_textR', 0,
			'child_textG', 1,
			'child_textB', 0
		)
	local _, petName, _, petFamily, petTalent = GetStablePetInfo(i);
	if petName ~= nil then
		tooltipLine:AddLine(
		'text', petName,
		'text2', petFamily..' ('..v.petTalent..')',
		'text2R', 1, 'text2G', 1, 'text2B', 0,
		'func', 'PetInfoFrame', 'arg1', self, 'arg2', v
		)              
	else
		toolTip:AddLine(
			'text', L["PI_NOPET"]
		)
	end
		-- add other pets if showing
		if self:IsHunter() and self.db.char.showActivePets then          
			self:AddActivePets()
		end   
		-- add stabled pets
		if self:IsHunter() and self.db.char.showStabledPets then
			self:AddStabledPets()
		end   
		-- add other demons if showing
		if self:IsWarlock() and self.db.char.showStabledPets then          
			self:AddNetheredDemons()
		end   
	end
end

function PetInfoFu:UpdateText()
	-- sound checks
	if self:IsHunter() then
		-- check for new talent points?
	end

	if  self:IsTextShown() then
		if self.db.char.showPetName then
			if UnitName("pet") ~= nil then
				local petName = UnitName("pet")
				local petIcon = GetPetIcon()
				local label = ''
				
				if petName ~= nil then  
				
					if petName == UNKNOWNOBJECT then
						petName = L["UO"]
						self:ScheduleEvent(self.UpdateText,2, self)
					end
				
					-- set pet icon for warlock
					if self:IsWarlock() then
						if (UnitCreatureFamily("pet") == L["Imp"]) then            petIcon = select(3, GetSpellInfo(688))   -- "Interface\\Icons\\Spell_Shadow_SummonImp"
						elseif (UnitCreatureFamily("pet") == L["Felguard"]) then   petIcon = select(3, GetSpellInfo(30146)) -- "Interface\\Icons\\Spell_Shadow_SummonFelGuard"
						elseif (UnitCreatureFamily("pet") == L["Felhunter"]) then  petIcon = select(3, GetSpellInfo(691))   -- "Interface\\Icons\\Spell_Shadow_SummonFelHunter"
						elseif (UnitCreatureFamily("pet") == L["Succubus"]) then   petIcon = select(3, GetSpellInfo(712))   -- "Interface\\Icons\\Spell_Shadow_SummonSuccubus"
						elseif (UnitCreatureFamily("pet") == L["Voidwalker"]) then petIcon = select(3, GetSpellInfo(697))   -- "Interface\\Icons\\Spell_Shadow_SummonVoidWalker"
						elseif (UnitCreatureFamily("pet") == L["Infernal"]) then   petIcon = select(3, GetSpellInfo(1122))  --"Interface\\Icons\\Spell_Shadow_SummonInfernal" 
						end
					end 
					
					if petIcon ~= nil then
						self:SetIcon(petIcon)        
					end
					-- show name and/or level
					if self.db.char.showPetName then                   
							label = petName
					end
			self:SetText(label)
			-- self:Print('PetInfoFu: Tooltip'..crayon:Red(UnitName("pet")))
				end
			else 
				-- no pet ui found
				self:SetText(L["PI_NOPET"])
				self:SetIcon(select(3, GetSpellInfo(2641)))
		-- self:Print('PetInfoFu: Tooltip'..crayon:Red('No pet UI'))
			end
		else
			if UnitName("pet") ~= nil then
				local petIcon = GetPetIcon()
				if petIcon ~= nil then
					self:SetIcon(petIcon)        
				end
			end
			-- self:Print('PetInfoFu: Tooltip'..crayon:Red('UnitName(pet) == nil'))
			self:SetText("")                    
		end
	end
end	

-- event handlers

function PetInfoFu:OnClick()
	if self:IsHunter() then
		if IsControlKeyDown() then
			-- initialise timer for show time til next level
			if self.db.char.showPetXP == 3 then
				self:ResetXPTimer()
			-- initialise XP counter for show kills til next level
			elseif self.db.char.showPetXP == 4 then
				local _, _, toLevelXP = self:PetInfo_GetPetExperience()                    
				prevXPToLevel = toLevelXP
			end            
		else
			self.db.char.hasPlayedSoundOnGainTP = false
			self.db.char.hasPlayedSoundOnSkillTrainable = false
			-- show talent frame, can't find how to open the actual pet tab
			ToggleTalentFrame()
		end
	end
end
